from aeronavx.core.network_intelligence import NetworkIntelligence
from aeronavx.models.airport import Airport


def _airport(code, airport_type, lat=0.0, lon=0.0):
    return Airport(
        id=None,
        ident=code,
        type=airport_type,
        name=f"{code} Airport",
        latitude_deg=lat,
        longitude_deg=lon,
        elevation_ft=None,
        continent=None,
        iso_country=None,
        iso_region=None,
        municipality=None,
        scheduled_service=True,
        gps_code=code,
        iata_code=code,
        local_code=None,
        home_link=None,
        wikipedia_link=None,
        keywords=None,
    )


def test_network_intelligence_ranking():
    airports = [
        _airport("AAA", "large_airport", 0.0, 0.0),
        _airport("BBB", "medium_airport", 0.1, 0.1),
        _airport("CCC", "small_airport", -0.1, -0.1),
    ]

    intelligence = NetworkIntelligence(airports, radius_km=500)
    ranked = intelligence.rank_airports()

    assert ranked[0].airport.iata_code == "AAA"
    assert ranked[0].hub_score >= ranked[1].hub_score >= ranked[2].hub_score



import pytest

from aeronavx.core import loader
from aeronavx.core.routing import shortest_path
from aeronavx.exceptions import RoutingError


def test_shortest_path_handles_duplicate_names(tmp_path):
    csv_content = "\n".join(
        [
            "id,ident,type,name,latitude_deg,longitude_deg,elevation_ft,continent,iso_country,iso_region,municipality,scheduled_service,gps_code,iata_code,local_code,home_link,wikipedia_link,keywords",
            "1,AAA1,small_airport,SameName,0.0,0.0,0,NA,XX,XX-1,Test City,no,AAA1,AAA,,,,",
            "2,AAA2,small_airport,SameName,0.0,10.0,0,NA,XX,XX-1,Test City,no,AAA2,AAB,,,,",
            "3,BBB1,small_airport,Destination,0.0,15.0,0,NA,XX,XX-1,Test City,no,BBB1,BBB,,,,",
        ]
    )
    csv_path = tmp_path / "airports.csv"
    csv_path.write_text(csv_content)

    loader.load_airports(data_path=csv_path, force_reload=True)

    with pytest.raises(RoutingError):
        shortest_path("AAA", "BBB", code_type="iata", max_leg_km=800)

    loader.clear_cache()


def test_route_distance_by_codes():
    from aeronavx.core.routing import route_distance_by_codes

    dist = route_distance_by_codes(["IST", "JFK"], code_type="iata", unit="km")

    assert dist > 1000



import os
from pathlib import Path
from typing import Optional


def _first_env(*names: str) -> Optional[str]:
    for name in names:
        value = os.getenv(name)
        if value:
            return value
    return None


HF_TOKEN = _first_env("HF_TOKEN", "HF_API_TOKEN", "HUGGINGFACE_HUB_TOKEN")
CACHE_DIR = Path(os.getenv("AERONAVX_CACHE", "~/.aeronavx")).expanduser()
EMBED_MODEL = os.getenv(
    "AERONAVX_EMBED_MODEL",
    "sentence-transformers/all-MiniLM-L6-v2",
)
USE_LOCAL_ONLY = os.getenv("AERONAVX_OFFLINE", "0") == "1"



import pytest

from aeronavx.hf.semantic_search import SemanticAirportSearch
from aeronavx.models.airport import Airport

np = pytest.importorskip("numpy")


class CountingEmbedder:
    def __init__(self, fail_on_bulk=False):
        self.bulk_calls = 0
        self.query_calls = 0
        self.fail_on_bulk = fail_on_bulk

    def encode(self, texts, normalize_embeddings=False):
        if isinstance(texts, str):
            texts = [texts]

        if len(texts) > 1:
            if self.fail_on_bulk:
                raise AssertionError("Bulk encode should be cached, but was called again.")
            self.bulk_calls += 1
        else:
            self.query_calls += 1

        vectors = []
        for text in texts:
            text_lower = text.lower()
            vec = np.array(
                [
                    float(text_lower.count("istanbul")),
                    float(text_lower.count("new")),
                    float(len(text_lower)),
                ],
                dtype=float,
            )
            vectors.append(vec)

        embeddings = np.vstack(vectors)

        if normalize_embeddings:
            norms = np.linalg.norm(embeddings, axis=1, keepdims=True)
            norms = np.where(norms == 0, 1.0, norms)
            embeddings = embeddings / norms

        return embeddings


def _sample_airports():
    return [
        Airport(
            id=1,
            ident="LTFM",
            type="large_airport",
            name="Istanbul Airport",
            latitude_deg=41.275278,
            longitude_deg=28.751944,
            elevation_ft=163,
            continent="AS",
            iso_country="TR",
            iso_region="TR-34",
            municipality="Istanbul",
            scheduled_service=True,
            gps_code="LTFM",
            iata_code="IST",
            local_code=None,
            home_link=None,
            wikipedia_link=None,
            keywords=None,
        ),
        Airport(
            id=2,
            ident="KJFK",
            type="large_airport",
            name="John F Kennedy International Airport",
            latitude_deg=40.639801,
            longitude_deg=-73.7789,
            elevation_ft=13,
            continent="NA",
            iso_country="US",
            iso_region="US-NY",
            municipality="New York",
            scheduled_service=True,
            gps_code="KJFK",
            iata_code="JFK",
            local_code=None,
            home_link=None,
            wikipedia_link=None,
            keywords=None,
        ),
    ]


def test_embedding_cache_reuse(tmp_path):
    airports = _sample_airports()

    embedder = CountingEmbedder()
    searcher = SemanticAirportSearch(
        airports,
        model_name="dummy-model",
        model=embedder,
        cache_dir=tmp_path,
    )

    results = searcher.search("Istanbul", top_k=1, return_format="list")
    assert results[0]["iata"] == "IST"
    assert embedder.bulk_calls == 1

    cached_files = list((tmp_path / "hf" / "semantic_search" / "dummy-model").glob("*.npz"))
    assert len(cached_files) == 1

    embedder_cached = CountingEmbedder(fail_on_bulk=True)
    searcher_cached = SemanticAirportSearch(
        airports,
        model_name="dummy-model",
        model=embedder_cached,
        cache_dir=tmp_path,
    )
    searcher_cached.search("Istanbul", top_k=1, return_format="list")


def test_clear_semantic_search_cache():
    import aeronavx

    aeronavx._semantic_engine = object()
    aeronavx.clear_semantic_search_cache()

    assert aeronavx._semantic_engine is None


def test_corrupted_cache_fallback(tmp_path):
    from aeronavx.hf.cache import embeddings_cache_path
    from aeronavx.hf.index import _fingerprint
    from aeronavx.hf.utils import airport_to_text

    airports = _sample_airports()
    texts = [airport_to_text(a) for a in airports]
    signature = _fingerprint(texts, "dummy-model")
    cache_path = embeddings_cache_path(tmp_path, "dummy-model", signature)
    cache_path.parent.mkdir(parents=True, exist_ok=True)
    cache_path.write_bytes(b"not-a-valid-npz")

    embedder = CountingEmbedder()
    searcher = SemanticAirportSearch(
        airports,
        model_name="dummy-model",
        model=embedder,
        cache_dir=tmp_path,
    )

    searcher.search("Istanbul", top_k=1, return_format="list")
    assert embedder.bulk_calls == 1


def test_partial_embeddings_fallback(tmp_path):
    from aeronavx.hf.cache import embeddings_cache_path
    from aeronavx.hf.index import _fingerprint
    from aeronavx.hf.utils import airport_to_text

    airports = _sample_airports()
    texts = [airport_to_text(a) for a in airports]
    signature = _fingerprint(texts, "dummy-model")
    cache_path = embeddings_cache_path(tmp_path, "dummy-model", signature)
    cache_path.parent.mkdir(parents=True, exist_ok=True)

    embeddings = np.zeros((1, 3), dtype=float)
    np.savez_compressed(
        cache_path,
        embeddings=embeddings,
        signature=signature,
        count=2,
        version="v1",
    )

    embedder = CountingEmbedder()
    searcher = SemanticAirportSearch(
        airports,
        model_name="dummy-model",
        model=embedder,
        cache_dir=tmp_path,
    )

    searcher.search("Istanbul", top_k=1, return_format="list")
    assert embedder.bulk_calls == 1



import pytest

from aeronavx.hf.semantic_search import SemanticAirportSearch
from aeronavx.models.airport import Airport

np = pytest.importorskip("numpy")


class DummyEmbedder:
    def encode(self, texts, normalize_embeddings=False):
        if isinstance(texts, str):
            texts = [texts]

        vectors = []
        for text in texts:
            text_lower = text.lower()
            vec = np.array(
                [
                    float(text_lower.count("istanbul")),
                    float(text_lower.count("new")),
                    float(len(text_lower)),
                ],
                dtype=float,
            )
            vectors.append(vec)

        embeddings = np.vstack(vectors)

        if normalize_embeddings:
            norms = np.linalg.norm(embeddings, axis=1, keepdims=True)
            norms = np.where(norms == 0, 1.0, norms)
            embeddings = embeddings / norms

        return embeddings


def test_semantic_search_ranks_expected_airport():
    airports = [
        Airport(
            id=1,
            ident="LTFM",
            type="large_airport",
            name="Istanbul Airport",
            latitude_deg=41.275278,
            longitude_deg=28.751944,
            elevation_ft=163,
            continent="AS",
            iso_country="TR",
            iso_region="TR-34",
            municipality="Istanbul",
            scheduled_service=True,
            gps_code="LTFM",
            iata_code="IST",
            local_code=None,
            home_link=None,
            wikipedia_link=None,
            keywords=None,
        ),
        Airport(
            id=2,
            ident="KJFK",
            type="large_airport",
            name="John F Kennedy International Airport",
            latitude_deg=40.639801,
            longitude_deg=-73.7789,
            elevation_ft=13,
            continent="NA",
            iso_country="US",
            iso_region="US-NY",
            municipality="New York",
            scheduled_service=True,
            gps_code="KJFK",
            iata_code="JFK",
            local_code=None,
            home_link=None,
            wikipedia_link=None,
            keywords=None,
        ),
        Airport(
            id=3,
            ident="RJTT",
            type="large_airport",
            name="Tokyo Haneda Airport",
            latitude_deg=35.552258,
            longitude_deg=139.779694,
            elevation_ft=21,
            continent="AS",
            iso_country="JP",
            iso_region="JP-13",
            municipality="Tokyo",
            scheduled_service=True,
            gps_code="RJTT",
            iata_code="HND",
            local_code=None,
            home_link=None,
            wikipedia_link=None,
            keywords=None,
        ),
    ]

    searcher = SemanticAirportSearch(
        airports,
        model_name="dummy",
        model=DummyEmbedder(),
        cache_dir=None,
    )

    results = searcher.search("Istanbul main airport", top_k=2, return_format="list")

    assert len(results) == 2
    assert results[0]["iata"] == "IST"


def test_return_format_auto_list_dataframe():
    airports = [
        Airport(
            id=1,
            ident="LTFM",
            type="large_airport",
            name="Istanbul Airport",
            latitude_deg=41.275278,
            longitude_deg=28.751944,
            elevation_ft=163,
            continent="AS",
            iso_country="TR",
            iso_region="TR-34",
            municipality="Istanbul",
            scheduled_service=True,
            gps_code="LTFM",
            iata_code="IST",
            local_code=None,
            home_link=None,
            wikipedia_link=None,
            keywords=None,
        )
    ]

    searcher = SemanticAirportSearch(
        airports,
        model_name="dummy",
        model=DummyEmbedder(),
        cache_dir=None,
    )

    auto_result = searcher.search("Istanbul airport", return_format="auto")
    list_result = searcher.search("Istanbul airport", return_format="list")

    assert isinstance(list_result, list)

    pandas = None
    try:
        import pandas as pd

        pandas = pd
    except ImportError:
        pass

    if pandas is None:
        assert isinstance(auto_result, list)
    else:
        df_result = searcher.search("Istanbul airport", return_format="dataframe")
        assert isinstance(auto_result, pandas.DataFrame)
        assert isinstance(df_result, pandas.DataFrame)



import argparse
import sys

from .. import __version__
from ..core.airports import get
from ..core.distance import distance
from ..core.emissions import estimate_co2_kg_by_codes
from ..core.emissions_advanced import (
    AircraftType,
    FuelType,
    calculate_flight_emissions,
    compare_saf_savings,
)
from ..core.network_intelligence import identify_global_hubs
from ..core.passenger_experience import calculate_jet_lag
from ..core.routing import estimate_flight_time_hours
from ..core.search import nearest_airports, search_airports_by_name
from ..core.synthetic_routes import generate_route
from ..exceptions import AeroNavXError


def cmd_distance(args):
    try:
        from_airport = get(args.from_code, code_type="auto")
        to_airport = get(args.to_code, code_type="auto")

        if from_airport is None:
            print(f"Error: Airport not found: {args.from_code}", file=sys.stderr)
            return 1

        if to_airport is None:
            print(f"Error: Airport not found: {args.to_code}", file=sys.stderr)
            return 1

        dist = distance(
            from_airport.latitude_deg,
            from_airport.longitude_deg,
            to_airport.latitude_deg,
            to_airport.longitude_deg,
            model=args.model,
            unit=args.unit,
        )

        print(f"{from_airport.name} ({args.from_code}) to {to_airport.name} ({args.to_code})")
        print(f"Distance: {dist:.2f} {args.unit}")

        return 0

    except AeroNavXError as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1


def cmd_nearest(args):
    try:
        airports = nearest_airports(args.lat, args.lon, n=args.n)

        if not airports:
            print("No airports found")
            return 0

        print(f"Nearest {len(airports)} airport(s) to ({args.lat}, {args.lon}):\n")

        for i, airport in enumerate(airports, 1):
            codes = []
            if airport.iata_code:
                codes.append(f"IATA: {airport.iata_code}")
            if airport.gps_code:
                codes.append(f"ICAO: {airport.gps_code}")

            code_str = f" ({', '.join(codes)})" if codes else ""

            dist_km = distance(
                args.lat, args.lon, airport.latitude_deg, airport.longitude_deg, unit="km"
            )

            print(f"{i}. {airport.name}{code_str}")
            print(f"   Distance: {dist_km:.2f} km")
            print(f"   Location: {airport.municipality}, {airport.iso_country}")
            print()

        return 0

    except AeroNavXError as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1


def cmd_search(args):
    try:
        airports = search_airports_by_name(args.name, limit=args.limit)

        if not airports:
            print(f"No airports found matching '{args.name}'")
            return 0

        print(f"Found {len(airports)} airport(s) matching '{args.name}':\n")

        for i, airport in enumerate(airports, 1):
            codes = []
            if airport.iata_code:
                codes.append(f"IATA: {airport.iata_code}")
            if airport.gps_code:
                codes.append(f"ICAO: {airport.gps_code}")

            code_str = f" ({', '.join(codes)})" if codes else ""

            print(f"{i}. {airport.name}{code_str}")
            print(f"   Type: {airport.type}")
            print(f"   Location: {airport.municipality}, {airport.iso_country}")
            print()

        return 0

    except AeroNavXError as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1


def cmd_emissions(args):
    try:
        co2_kg = estimate_co2_kg_by_codes(
            args.from_code, args.to_code, code_type="auto", model=args.model
        )

        from_airport = get(args.from_code, code_type="auto")
        to_airport = get(args.to_code, code_type="auto")

        print(f"{from_airport.name} ({args.from_code}) to {to_airport.name} ({args.to_code})")
        print(f"Estimated CO2 emissions: {co2_kg:.2f} kg per passenger")

        return 0

    except AeroNavXError as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1


def cmd_flight_time(args):
    try:
        from_airport = get(args.from_code, code_type="auto")
        to_airport = get(args.to_code, code_type="auto")

        if from_airport is None:
            print(f"Error: Airport not found: {args.from_code}", file=sys.stderr)
            return 1

        if to_airport is None:
            print(f"Error: Airport not found: {args.to_code}", file=sys.stderr)
            return 1

        time_hours = estimate_flight_time_hours(from_airport, to_airport, speed_kts=args.speed_kts)

        hours = int(time_hours)
        minutes = int((time_hours - hours) * 60)

        print(f"{from_airport.name} ({args.from_code}) to {to_airport.name} ({args.to_code})")
        print(f"Estimated flight time: {hours}h {minutes}m (at {args.speed_kts} knots)")

        return 0

    except AeroNavXError as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1


def cmd_semantic_search(args):
    try:
        from .. import semantic_search

        results = semantic_search(args.query, top_k=args.top_k, return_format="list")

        if not results:
            print("No results found")
            return 0

        print(f"Semantic search results for '{args.query}':\n")
        for idx, row in enumerate(results, 1):
            code = row.get("iata") or row.get("icao") or "---"
            name = row.get("name") or "Unknown"
            score = row.get("score", 0.0)
            print(f"{idx}. {code} - {name} (score={score:.3f})")

        return 0
    except ImportError as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1


def cmd_jet_lag(args):
    try:
        origin = get(args.from_code, code_type="auto")
        destination = get(args.to_code, code_type="auto")

        if origin is None or destination is None:
            print("Error: Airport not found", file=sys.stderr)
            return 1

        result = calculate_jet_lag(origin, destination, age=args.age)

        print(f"{origin.name} ({args.from_code}) -> {destination.name} ({args.to_code})")
        print(f"Timezone difference: {result.timezone_difference_hours:.1f} hours")
        print(f"Direction: {result.direction.value}")
        print(f"Severity: {result.severity.value}")
        print(f"Estimated recovery: {result.estimated_recovery_days:.1f} days")

        return 0
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1


def cmd_hubs(args):
    try:
        hubs = identify_global_hubs(
            top_n=args.top_n, scheduled_service_only=not args.include_unscheduled
        )

        print(f"Top {len(hubs)} hubs:")
        for hub in hubs:
            print(f"{hub.airport.iata_code or '---'} - {hub.airport.name}")
            print(f"  Hub Score: {hub.hub_score:.2f}")
            print(f"  Connectivity: {hub.connectivity_score:.2f}")

        return 0
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1


def cmd_emissions_advanced(args):
    try:
        aircraft = AircraftType(args.aircraft_type)
        fuel = FuelType(args.fuel_type)

        emissions = calculate_flight_emissions(
            args.from_code,
            args.to_code,
            aircraft_type=aircraft,
            fuel_type=fuel,
            load_factor=args.load_factor,
            code_type="auto",
        )

        print(f"Route: {args.from_code} -> {args.to_code}")
        print(f"Distance: {emissions.distance_km:.1f} km")
        print(f"Fuel: {emissions.fuel_kg:.1f} kg")
        print(f"Total CO2: {emissions.total_co2_kg:.1f} kg")
        print(f"Per passenger: {emissions.co2_per_passenger_kg:.2f} kg")

        if args.saf_percent > 0:
            savings = compare_saf_savings(
                args.from_code,
                args.to_code,
                saf_percentage=args.saf_percent,
                aircraft_type=aircraft,
                load_factor=args.load_factor,
                code_type="auto",
            )
            print(
                f"SAF {args.saf_percent:.0f}% reduction: "
                f"{savings.co2_reduction_kg:.1f} kg ({savings.reduction_percentage:.1f}%)"
            )

        return 0
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1


def cmd_synthetic_route(args):
    try:
        route = generate_route(
            args.from_code,
            args.to_code,
            num_waypoints=args.waypoints,
            speed_kts=args.speed_kts,
        )

        print(f"Route: {route.origin.name} -> {route.destination.name}")
        print(f"Total distance: {route.total_distance_km:.1f} km")
        print(f"Estimated time: {route.total_time_hours:.2f} hours")
        print(f"Waypoints: {len(route.waypoints)}")

        preview = route.waypoints[: min(5, len(route.waypoints))]
        for wp in preview:
            print(f"  {wp.name}: ({wp.latitude:.2f}, {wp.longitude:.2f})")

        return 0
    except Exception as e:
        print(f"Error: {e}", file=sys.stderr)
        return 1


def main():
    parser = argparse.ArgumentParser(prog="aeronavx", description="AeroNavX - AI aviation SDK CLI")
    parser.add_argument(
        "--version",
        action="version",
        version=f"aeronavx {__version__}",
    )

    subparsers = parser.add_subparsers(dest="command", help="Available commands")

    distance_parser = subparsers.add_parser("distance", help="Calculate distance between airports")
    distance_parser.add_argument(
        "--from", dest="from_code", required=True, help="Origin airport code"
    )
    distance_parser.add_argument(
        "--to", dest="to_code", required=True, help="Destination airport code"
    )
    distance_parser.add_argument(
        "--unit", choices=["km", "mi", "nmi"], default="km", help="Distance unit"
    )
    distance_parser.add_argument(
        "--model",
        choices=["haversine", "slc", "vincenty"],
        default="haversine",
        help="Distance model",
    )

    nearest_parser = subparsers.add_parser("nearest", help="Find nearest airports to coordinates")
    nearest_parser.add_argument("--lat", type=float, required=True, help="Latitude")
    nearest_parser.add_argument("--lon", type=float, required=True, help="Longitude")
    nearest_parser.add_argument("--n", type=int, default=5, help="Number of airports to return")

    search_parser = subparsers.add_parser("search", help="Search airports by name")
    search_parser.add_argument("--name", required=True, help="Search query")
    search_parser.add_argument("--limit", type=int, default=10, help="Maximum results")

    emissions_parser = subparsers.add_parser("emissions", help="Estimate CO2 emissions")
    emissions_parser.add_argument(
        "--from", dest="from_code", required=True, help="Origin airport code"
    )
    emissions_parser.add_argument(
        "--to", dest="to_code", required=True, help="Destination airport code"
    )
    emissions_parser.add_argument(
        "--model",
        choices=["haversine", "slc", "vincenty"],
        default="haversine",
        help="Distance model",
    )

    emissions_adv_parser = subparsers.add_parser(
        "emissions-advanced", help="Advanced emissions model"
    )
    emissions_adv_parser.add_argument(
        "--from", dest="from_code", required=True, help="Origin airport code"
    )
    emissions_adv_parser.add_argument(
        "--to", dest="to_code", required=True, help="Destination airport code"
    )
    emissions_adv_parser.add_argument(
        "--aircraft-type",
        default="narrow_body",
        choices=[t.value for t in AircraftType],
        help="Aircraft category",
    )
    emissions_adv_parser.add_argument(
        "--fuel-type",
        default="jet_a1",
        choices=[t.value for t in FuelType],
        help="Fuel type",
    )
    emissions_adv_parser.add_argument(
        "--load-factor", type=float, default=0.85, help="Load factor (0-1)"
    )
    emissions_adv_parser.add_argument(
        "--saf-percent", type=float, default=0.0, help="SAF blend percentage"
    )

    time_parser = subparsers.add_parser("flight-time", help="Estimate flight time")
    time_parser.add_argument("--from", dest="from_code", required=True, help="Origin airport code")
    time_parser.add_argument("--to", dest="to_code", required=True, help="Destination airport code")
    time_parser.add_argument("--speed-kts", type=float, default=450.0, help="Cruise speed in knots")

    semantic_parser = subparsers.add_parser("semantic-search", help="AI semantic airport search")
    semantic_parser.add_argument("--query", required=True, help="Search query")
    semantic_parser.add_argument("--top-k", type=int, default=5, help="Number of results")

    jet_lag_parser = subparsers.add_parser("jet-lag", help="Jet lag analysis")
    jet_lag_parser.add_argument(
        "--from", dest="from_code", required=True, help="Origin airport code"
    )
    jet_lag_parser.add_argument(
        "--to", dest="to_code", required=True, help="Destination airport code"
    )
    jet_lag_parser.add_argument("--age", type=int, default=30, help="Passenger age")

    hubs_parser = subparsers.add_parser("hubs", help="Identify global hubs")
    hubs_parser.add_argument("--top-n", type=int, default=10, help="Number of hubs")
    hubs_parser.add_argument(
        "--include-unscheduled",
        action="store_true",
        help="Include airports without scheduled service",
    )

    route_parser = subparsers.add_parser("synthetic-route", help="Generate synthetic route")
    route_parser.add_argument("--from", dest="from_code", required=True, help="Origin airport code")
    route_parser.add_argument(
        "--to", dest="to_code", required=True, help="Destination airport code"
    )
    route_parser.add_argument("--waypoints", type=int, default=12, help="Number of waypoints")
    route_parser.add_argument(
        "--speed-kts", type=float, default=450.0, help="Cruise speed in knots"
    )

    args = parser.parse_args()

    if not args.command:
        parser.print_help()
        return 1

    if args.command == "distance":
        return cmd_distance(args)
    elif args.command == "nearest":
        return cmd_nearest(args)
    elif args.command == "search":
        return cmd_search(args)
    elif args.command == "emissions":
        return cmd_emissions(args)
    elif args.command == "emissions-advanced":
        return cmd_emissions_advanced(args)
    elif args.command == "flight-time":
        return cmd_flight_time(args)
    elif args.command == "semantic-search":
        return cmd_semantic_search(args)
    elif args.command == "jet-lag":
        return cmd_jet_lag(args)
    elif args.command == "hubs":
        return cmd_hubs(args)
    elif args.command == "synthetic-route":
        return cmd_synthetic_route(args)
    else:
        parser.print_help()
        return 1


if __name__ == "__main__":
    sys.exit(main())



from __future__ import annotations

import hashlib
from pathlib import Path
from typing import Any, Optional

from .cache import CACHE_VERSION, embeddings_cache_path, load_embeddings, save_embeddings
from .utils import TEXT_VERSION, airport_to_record, airport_to_text, coerce_airports


def _normalize_embeddings(embeddings):
    import numpy as np

    norms = np.linalg.norm(embeddings, axis=1, keepdims=True)
    norms = np.where(norms == 0, 1.0, norms)
    return embeddings / norms


def _encode(model: Any, texts: list[str]):
    try:
        embeddings = model.encode(texts, normalize_embeddings=True)
    except TypeError:
        embeddings = model.encode(texts)
    return embeddings


def _fingerprint(texts: list[str], model_id: str) -> str:
    hasher = hashlib.sha256()
    hasher.update(CACHE_VERSION.encode("utf-8"))
    hasher.update(TEXT_VERSION.encode("utf-8"))
    hasher.update(model_id.encode("utf-8"))
    for text in texts:
        hasher.update(text.encode("utf-8"))
        hasher.update(b"\0")
    return hasher.hexdigest()


class AirportIndex:
    def __init__(
        self,
        airports: Any,
        model: Any,
        model_id: str,
        cache_dir: Optional[Path] = None,
        use_faiss: bool = True,
    ) -> None:
        self.model = model
        self.model_id = model_id
        self.cache_dir = cache_dir
        self.use_faiss = use_faiss

        airports_list = coerce_airports(airports)
        self.records = [airport_to_record(a) for a in airports_list]
        self.texts = [airport_to_text(a) for a in airports_list]

        self.embeddings = self._load_or_build_embeddings()
        self._faiss_index = self._build_faiss_index()

    def _load_or_build_embeddings(self):
        import numpy as np

        if not self.texts:
            return np.zeros((0, 0), dtype=float)

        signature = _fingerprint(self.texts, self.model_id)
        cached = None
        if self.cache_dir is not None:
            cache_path = embeddings_cache_path(self.cache_dir, self.model_id, signature)
            cached = load_embeddings(cache_path)

        if cached:
            embeddings, sig, count, version = cached
            if (
                sig == signature
                and count == len(self.texts)
                and embeddings.shape[0] == len(self.texts)
                and version == CACHE_VERSION
            ):
                return embeddings

        embeddings = _encode(self.model, self.texts)
        embeddings = np.asarray(embeddings)
        embeddings = _normalize_embeddings(embeddings)

        if self.cache_dir is not None:
            cache_path = embeddings_cache_path(self.cache_dir, self.model_id, signature)
            save_embeddings(cache_path, embeddings, signature)

        return embeddings

    def _build_faiss_index(self):
        if not self.use_faiss:
            return None

        try:
            import faiss
        except Exception:
            return None

        if self.embeddings is None or len(self.embeddings) == 0:
            return None

        dims = int(self.embeddings.shape[1])
        if dims == 0:
            return None
        index = faiss.IndexFlatIP(dims)
        index.add(self.embeddings.astype("float32"))
        return index

    def search(self, query: str, top_k: int = 5) -> list[dict[str, Any]]:
        import numpy as np

        if top_k <= 0 or len(self.embeddings) == 0:
            return []

        query_embeddings = _encode(self.model, [query])
        query_embeddings = np.asarray(query_embeddings)
        if query_embeddings.ndim == 2:
            query_embeddings = query_embeddings[0]
        query_embeddings = query_embeddings.reshape(1, -1)
        query_embeddings = _normalize_embeddings(query_embeddings)[0]

        if self._faiss_index is not None:
            top_k = min(top_k, len(self.records))
            scores, indices = self._faiss_index.search(
                query_embeddings.astype("float32"),
                top_k,
            )
            scores = scores[0]
            indices = indices[0]
        else:
            scores = np.dot(self.embeddings, query_embeddings)
            top_k = min(top_k, len(scores))
            if top_k == 0:
                return []
            if top_k == len(scores):
                indices = np.argsort(scores)[::-1]
            else:
                indices = np.argpartition(-scores, top_k - 1)[:top_k]
                indices = indices[np.argsort(scores[indices])[::-1]]

        results = []
        for idx in indices:
            if idx < 0:
                continue
            record = dict(self.records[idx])
            record["score"] = float(scores[idx])
            results.append(record)

        return results



from fastapi import FastAPI, HTTPException, Query

from ..core.airports import get
from ..core.distance import distance
from ..core.emissions import estimate_co2_kg_by_codes
from ..core.emissions_advanced import (
    AircraftType,
    FuelType,
    calculate_flight_emissions,
    compare_saf_savings,
)
from ..core.network_intelligence import identify_global_hubs
from ..core.passenger_experience import calculate_jet_lag
from ..core.routing import estimate_flight_time_hours
from ..core.search import nearest_airports, search_airports_by_name
from ..core.synthetic_routes import generate_route
from ..exceptions import AeroNavXError

app = FastAPI(title="AeroNavX API", description="AI aviation SDK API", version="3.0.1")


_semantic_engine = None


def _get_semantic_engine():
    global _semantic_engine
    if _semantic_engine is None:
        try:
            from ..hf.semantic_search import SemanticAirportSearch
        except ImportError as exc:
            raise HTTPException(
                status_code=501,
                detail=(
                    "Semantic search requires the HF extra. "
                    "Install with `pip install aeronavx[hf]`."
                ),
            ) from exc
        from ..core.loader import get_all_airports

        _semantic_engine = SemanticAirportSearch(get_all_airports())
    return _semantic_engine


@app.get("/health")
async def health():
    return {"status": "ok", "service": "AeroNavX API"}


@app.get("/airport/{code}")
async def get_airport(code: str, code_type: str = Query("auto", regex="^(iata|icao|auto)$")):
    try:
        airport = get(code, code_type=code_type)

        if airport is None:
            raise HTTPException(status_code=404, detail=f"Airport not found: {code}")

        return airport.as_dict()

    except AeroNavXError as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/distance")
async def calculate_distance(
    from_code: str = Query(..., alias="from"),
    to_code: str = Query(..., alias="to"),
    code_type: str = Query("auto", regex="^(iata|icao|auto)$"),
    model: str = Query("haversine", regex="^(haversine|slc|vincenty)$"),
    unit: str = Query("km", regex="^(km|mi|nmi)$"),
):
    try:
        from_airport = get(from_code, code_type=code_type)
        to_airport = get(to_code, code_type=code_type)

        if from_airport is None:
            raise HTTPException(status_code=404, detail=f"Origin airport not found: {from_code}")

        if to_airport is None:
            raise HTTPException(status_code=404, detail=f"Destination airport not found: {to_code}")

        dist = distance(
            from_airport.latitude_deg,
            from_airport.longitude_deg,
            to_airport.latitude_deg,
            to_airport.longitude_deg,
            model=model,
            unit=unit,
        )

        return {
            "from": from_airport.as_dict(),
            "to": to_airport.as_dict(),
            "distance": dist,
            "unit": unit,
            "model": model,
        }

    except AeroNavXError as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/nearest")
async def find_nearest(
    lat: float = Query(..., ge=-90, le=90),
    lon: float = Query(..., ge=-180, le=180),
    n: int = Query(5, ge=1, le=100),
):
    try:
        airports = nearest_airports(lat, lon, n=n)

        return {
            "query": {"lat": lat, "lon": lon},
            "count": len(airports),
            "airports": [a.as_dict() for a in airports],
        }

    except AeroNavXError as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/search")
async def search(q: str = Query(..., min_length=1), limit: int = Query(20, ge=1, le=100)):
    try:
        airports = search_airports_by_name(q, limit=limit)

        return {"query": q, "count": len(airports), "airports": [a.as_dict() for a in airports]}

    except AeroNavXError as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/flight-time")
async def flight_time(
    from_code: str = Query(..., alias="from"),
    to_code: str = Query(..., alias="to"),
    speed_kts: float = Query(450.0, ge=100, le=1000),
):
    try:
        from_airport = get(from_code, code_type="auto")
        to_airport = get(to_code, code_type="auto")

        if from_airport is None:
            raise HTTPException(status_code=404, detail=f"Origin airport not found: {from_code}")

        if to_airport is None:
            raise HTTPException(status_code=404, detail=f"Destination airport not found: {to_code}")

        time_hours = estimate_flight_time_hours(from_airport, to_airport, speed_kts=speed_kts)

        hours = int(time_hours)
        minutes = int((time_hours - hours) * 60)

        return {
            "from": from_airport.as_dict(),
            "to": to_airport.as_dict(),
            "speed_kts": speed_kts,
            "time_hours": time_hours,
            "time_formatted": f"{hours}h {minutes}m",
        }

    except AeroNavXError as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/emissions")
async def emissions(
    from_code: str = Query(..., alias="from"),
    to_code: str = Query(..., alias="to"),
    model: str = Query("haversine", regex="^(haversine|slc|vincenty)$"),
):
    try:
        co2_kg = estimate_co2_kg_by_codes(from_code, to_code, code_type="auto", model=model)

        from_airport = get(from_code, code_type="auto")
        to_airport = get(to_code, code_type="auto")

        return {
            "from": from_airport.as_dict(),
            "to": to_airport.as_dict(),
            "co2_kg_per_passenger": co2_kg,
            "model": model,
        }

    except AeroNavXError as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/semantic-search")
async def semantic_search(
    q: str = Query(..., min_length=1),
    top_k: int = Query(5, ge=1, le=50),
):
    try:
        engine = _get_semantic_engine()
        results = engine.search(q, top_k=top_k, return_format="list")
        return {"query": q, "count": len(results), "results": results}
    except HTTPException:
        raise
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/jet-lag")
async def jet_lag(
    from_code: str = Query(..., alias="from"),
    to_code: str = Query(..., alias="to"),
    age: int = Query(30, ge=1, le=120),
):
    try:
        origin = get(from_code, code_type="auto")
        destination = get(to_code, code_type="auto")

        if origin is None or destination is None:
            raise HTTPException(status_code=404, detail="Airport not found")

        result = calculate_jet_lag(origin, destination, age=age)
        return {
            "from": origin.as_dict(),
            "to": destination.as_dict(),
            "timezone_difference_hours": result.timezone_difference_hours,
            "direction": result.direction.value,
            "severity": result.severity.value,
            "estimated_recovery_days": result.estimated_recovery_days,
        }
    except HTTPException:
        raise
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/hubs")
async def hubs(
    top_n: int = Query(10, ge=1, le=50),
    include_unscheduled: bool = Query(False),
):
    try:
        results = identify_global_hubs(top_n=top_n, scheduled_service_only=not include_unscheduled)
        return {
            "count": len(results),
            "hubs": [
                {
                    "airport": hub.airport.as_dict(),
                    "hub_score": hub.hub_score,
                    "connectivity_score": hub.connectivity_score,
                    "type_score": hub.type_score,
                }
                for hub in results
            ],
        }
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/emissions-advanced")
async def emissions_advanced(
    from_code: str = Query(..., alias="from"),
    to_code: str = Query(..., alias="to"),
    aircraft_type: str = Query("narrow_body"),
    fuel_type: str = Query("jet_a1"),
    load_factor: float = Query(0.85, ge=0.1, le=1.0),
    saf_percent: float = Query(0.0, ge=0.0, le=100.0),
):
    try:
        emissions = calculate_flight_emissions(
            from_code,
            to_code,
            aircraft_type=AircraftType(aircraft_type),
            fuel_type=FuelType(fuel_type),
            load_factor=load_factor,
            code_type="auto",
        )
        response = {
            "from": from_code,
            "to": to_code,
            "distance_km": emissions.distance_km,
            "fuel_kg": emissions.fuel_kg,
            "total_co2_kg": emissions.total_co2_kg,
            "co2_per_passenger_kg": emissions.co2_per_passenger_kg,
            "aircraft_type": emissions.aircraft_type.value,
            "fuel_type": emissions.fuel_type.value,
            "load_factor": emissions.load_factor,
        }
        if saf_percent > 0:
            savings = compare_saf_savings(
                from_code,
                to_code,
                saf_percentage=saf_percent,
                aircraft_type=AircraftType(aircraft_type),
                load_factor=load_factor,
                code_type="auto",
            )
            response["saf"] = {
                "base_co2_kg": savings.base_co2_kg,
                "saf_co2_kg": savings.saf_co2_kg,
                "co2_reduction_kg": savings.co2_reduction_kg,
                "reduction_percentage": savings.reduction_percentage,
            }
        return response
    except ValueError as e:
        raise HTTPException(status_code=400, detail=str(e))
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


@app.get("/synthetic-route")
async def synthetic_route(
    from_code: str = Query(..., alias="from"),
    to_code: str = Query(..., alias="to"),
    waypoints: int = Query(12, ge=0, le=100),
    speed_kts: float = Query(450.0, ge=100, le=1000),
):
    try:
        route = generate_route(
            from_code,
            to_code,
            num_waypoints=waypoints,
            speed_kts=speed_kts,
        )
        return {
            "from": route.origin.as_dict(),
            "to": route.destination.as_dict(),
            "total_distance_km": route.total_distance_km,
            "total_time_hours": route.total_time_hours,
            "waypoints": [
                {
                    "name": wp.name,
                    "latitude": wp.latitude,
                    "longitude": wp.longitude,
                    "distance_from_origin_km": wp.distance_from_origin_km,
                }
                for wp in route.waypoints
            ],
        }
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e))


def run_server(host: str = "0.0.0.0", port: int = 8000):
    import uvicorn

    uvicorn.run(app, host=host, port=port)


if __name__ == "__main__":
    run_server()



import importlib
import sys
from types import SimpleNamespace

import pytest


def test_offline_mode_missing_cache_raises(monkeypatch, tmp_path):
    class DummySentenceTransformer:
        def __init__(self, model_name, **kwargs):
            assert kwargs.get("local_files_only") is True
            raise OSError("Model not found in cache")

    dummy_module = SimpleNamespace(SentenceTransformer=DummySentenceTransformer)
    monkeypatch.setitem(sys.modules, "sentence_transformers", dummy_module)
    if "aeronavx.hf.semantic_search" in sys.modules:
        del sys.modules["aeronavx.hf.semantic_search"]

    import aeronavx.hf.semantic_search as semantic_search

    importlib.reload(semantic_search)

    with pytest.raises(RuntimeError, match="Offline mode enabled"):
        semantic_search.SemanticAirportSearch([], cache_dir=tmp_path, local_only=True)


def test_invalid_token_error(monkeypatch, tmp_path):
    class DummySentenceTransformer:
        def __init__(self, model_name, **kwargs):
            raise RuntimeError("401 Unauthorized")

    dummy_module = SimpleNamespace(SentenceTransformer=DummySentenceTransformer)
    monkeypatch.setitem(sys.modules, "sentence_transformers", dummy_module)

    if "aeronavx.hf.semantic_search" in sys.modules:
        del sys.modules["aeronavx.hf.semantic_search"]

    import aeronavx.hf.semantic_search as semantic_search

    importlib.reload(semantic_search)

    with pytest.raises(RuntimeError, match="token is invalid"):
        semantic_search.SemanticAirportSearch([], cache_dir=tmp_path, token="invalid-token")


def test_network_error(monkeypatch, tmp_path):
    class DummySentenceTransformer:
        def __init__(self, model_name, **kwargs):
            raise RuntimeError("Connection timed out")

    dummy_module = SimpleNamespace(SentenceTransformer=DummySentenceTransformer)
    monkeypatch.setitem(sys.modules, "sentence_transformers", dummy_module)

    if "aeronavx.hf.semantic_search" in sys.modules:
        del sys.modules["aeronavx.hf.semantic_search"]

    import aeronavx.hf.semantic_search as semantic_search

    importlib.reload(semantic_search)

    with pytest.raises(RuntimeError, match="Unable to download model"):
        semantic_search.SemanticAirportSearch([], cache_dir=tmp_path)



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum

from ..core.distance import distance
from ..core.loader import get_airport_by_iata, get_airport_by_icao
from ..exceptions import RoutingError
from ..utils.constants import DEFAULT_CO2_KG_PER_PAX_KM


class AircraftType(Enum):
    REGIONAL = "regional"
    NARROW_BODY = "narrow_body"
    WIDE_BODY = "wide_body"
    BUSINESS_JET = "business_jet"


class FuelType(Enum):
    JET_A1 = "jet_a1"
    SAF = "saf"


AIRCRAFT_SEATS = {
    AircraftType.REGIONAL: 90,
    AircraftType.NARROW_BODY: 180,
    AircraftType.WIDE_BODY: 300,
    AircraftType.BUSINESS_JET: 8,
}

FUEL_BURN_KG_PER_KM = {
    AircraftType.REGIONAL: 1200 / 1000.0,
    AircraftType.NARROW_BODY: 2500 / 1000.0,
    AircraftType.WIDE_BODY: 5200 / 1000.0,
    AircraftType.BUSINESS_JET: 900 / 1000.0,
}

FUEL_CO2_MULTIPLIER = {
    FuelType.JET_A1: 1.0,
    FuelType.SAF: 0.5,
}

CO2_PER_KG_FUEL = 3.16


@dataclass(frozen=True)
class EmissionsResult:
    distance_km: float
    fuel_kg: float
    total_co2_kg: float
    co2_per_passenger_kg: float
    load_factor: float
    aircraft_type: AircraftType
    fuel_type: FuelType


@dataclass(frozen=True)
class SafSavings:
    base_co2_kg: float
    saf_co2_kg: float
    co2_reduction_kg: float
    reduction_percentage: float


def _resolve_airports(from_code: str, to_code: str, code_type: str) -> tuple:
    if code_type == "iata":
        from_airport = get_airport_by_iata(from_code)
        to_airport = get_airport_by_iata(to_code)
    elif code_type == "icao":
        from_airport = get_airport_by_icao(from_code)
        to_airport = get_airport_by_icao(to_code)
    else:
        from_airport = get_airport_by_iata(from_code) or get_airport_by_icao(from_code)
        to_airport = get_airport_by_iata(to_code) or get_airport_by_icao(to_code)

    if from_airport is None:
        raise RoutingError(f"Origin airport not found: {from_code}")
    if to_airport is None:
        raise RoutingError(f"Destination airport not found: {to_code}")

    return from_airport, to_airport


def calculate_flight_emissions(
    from_code: str,
    to_code: str,
    aircraft_type: AircraftType = AircraftType.NARROW_BODY,
    fuel_type: FuelType = FuelType.JET_A1,
    load_factor: float = 0.85,
    code_type: str = "iata",
    model: str = "haversine",
) -> EmissionsResult:
    if not 0 < load_factor <= 1.0:
        raise ValueError("load_factor must be in (0, 1].")

    from_airport, to_airport = _resolve_airports(from_code, to_code, code_type)

    dist_km = distance(
        from_airport.latitude_deg,
        from_airport.longitude_deg,
        to_airport.latitude_deg,
        to_airport.longitude_deg,
        model=model,
        unit="km",
    )

    seats = AIRCRAFT_SEATS.get(aircraft_type, 180)
    passengers = max(int(seats * load_factor), 1)

    fuel_burn = FUEL_BURN_KG_PER_KM.get(aircraft_type, 2.5)
    fuel_kg = dist_km * fuel_burn

    co2_multiplier = FUEL_CO2_MULTIPLIER.get(fuel_type, 1.0)
    total_co2 = fuel_kg * CO2_PER_KG_FUEL * co2_multiplier
    per_passenger = total_co2 / passengers

    baseline_per_pax = dist_km * DEFAULT_CO2_KG_PER_PAX_KM
    per_passenger = max(per_passenger, baseline_per_pax * 0.5)

    return EmissionsResult(
        distance_km=dist_km,
        fuel_kg=fuel_kg,
        total_co2_kg=total_co2,
        co2_per_passenger_kg=per_passenger,
        load_factor=load_factor,
        aircraft_type=aircraft_type,
        fuel_type=fuel_type,
    )


def compare_saf_savings(
    from_code: str,
    to_code: str,
    saf_percentage: float = 50.0,
    aircraft_type: AircraftType = AircraftType.NARROW_BODY,
    load_factor: float = 0.85,
    code_type: str = "iata",
    model: str = "haversine",
) -> SafSavings:
    if not 0 <= saf_percentage <= 100:
        raise ValueError("saf_percentage must be between 0 and 100.")

    base = calculate_flight_emissions(
        from_code,
        to_code,
        aircraft_type=aircraft_type,
        fuel_type=FuelType.JET_A1,
        load_factor=load_factor,
        code_type=code_type,
        model=model,
    )

    saf_multiplier = 1.0 - 0.7 * (saf_percentage / 100.0)
    saf_co2 = base.total_co2_kg * saf_multiplier
    reduction = base.total_co2_kg - saf_co2
    reduction_pct = (reduction / base.total_co2_kg) * 100 if base.total_co2_kg else 0.0

    return SafSavings(
        base_co2_kg=base.total_co2_kg,
        saf_co2_kg=saf_co2,
        co2_reduction_kg=reduction,
        reduction_percentage=reduction_pct,
    )



from __future__ import annotations

from collections.abc import Sequence
from dataclasses import dataclass
from typing import Union

from ..core.airports import get as get_airport
from ..core.distance import distance
from ..core.geodesy import great_circle_path
from ..exceptions import RoutingError
from ..models.airport import Airport
from ..utils.constants import DEFAULT_CRUISE_SPEED_KTS


@dataclass(frozen=True)
class Waypoint:
    name: str
    latitude: float
    longitude: float
    distance_from_origin_km: float


@dataclass(frozen=True)
class SyntheticRoute:
    origin: Airport
    destination: Airport
    waypoints: list[Waypoint]
    total_distance_km: float
    total_time_hours: float


def _resolve_airport(value: Union[str, Airport]) -> Airport:
    if isinstance(value, Airport):
        return value

    airport = get_airport(value, code_type="auto")
    if airport is None:
        raise RoutingError(f"Airport not found: {value}")
    return airport


def generate_route(
    origin: Union[str, Airport],
    destination: Union[str, Airport],
    num_waypoints: int = 12,
    model: str = "haversine",
    speed_kts: float = DEFAULT_CRUISE_SPEED_KTS,
) -> SyntheticRoute:
    if num_waypoints < 0:
        raise ValueError("num_waypoints must be >= 0")

    origin_airport = _resolve_airport(origin)
    destination_airport = _resolve_airport(destination)

    num_points = max(num_waypoints + 2, 2)
    path_points = great_circle_path(
        origin_airport.latitude_deg,
        origin_airport.longitude_deg,
        destination_airport.latitude_deg,
        destination_airport.longitude_deg,
        num_points=num_points,
    )

    waypoints: list[Waypoint] = []
    for idx, (lat, lon) in enumerate(path_points[1:-1], start=1):
        dist_km = distance(
            origin_airport.latitude_deg,
            origin_airport.longitude_deg,
            lat,
            lon,
            model=model,
            unit="km",
        )
        waypoints.append(
            Waypoint(
                name=f"WP{idx}",
                latitude=lat,
                longitude=lon,
                distance_from_origin_km=dist_km,
            )
        )

    total_distance_km = 0.0
    for start, end in zip(path_points[:-1], path_points[1:]):
        total_distance_km += distance(
            start[0],
            start[1],
            end[0],
            end[1],
            model=model,
            unit="km",
        )

    speed_kmh = speed_kts * 1.852
    total_time_hours = total_distance_km / speed_kmh if speed_kmh > 0 else 0.0

    return SyntheticRoute(
        origin=origin_airport,
        destination=destination_airport,
        waypoints=waypoints,
        total_distance_km=total_distance_km,
        total_time_hours=total_time_hours,
    )


def generate_route_by_codes(
    codes: Sequence[str],
    model: str = "haversine",
    speed_kts: float = DEFAULT_CRUISE_SPEED_KTS,
) -> SyntheticRoute:
    if len(codes) < 2:
        raise RoutingError("At least two airport codes are required.")

    origin = _resolve_airport(codes[0])
    destination = _resolve_airport(codes[-1])

    return generate_route(origin, destination, model=model, speed_kts=speed_kts)



from datetime import datetime
from typing import Optional
from zoneinfo import ZoneInfo, ZoneInfoNotFoundError

from ..core.loader import get_airport_by_iata, get_airport_by_icao
from ..models.airport import Airport
from ..utils.logging import get_logger

logger = get_logger()


try:
    from timezonefinder import TimezoneFinder

    _tz_finder = TimezoneFinder()
    HAS_TIMEZONEFINDER = True
except ImportError:
    _tz_finder = None
    HAS_TIMEZONEFINDER = False
    logger.warning("timezonefinder not installed, timezone functionality will be limited")


def get_timezone_for_airport(airport: Airport) -> Optional[str]:
    if not HAS_TIMEZONEFINDER:
        return None

    try:
        tz_name = _tz_finder.timezone_at(lat=airport.latitude_deg, lng=airport.longitude_deg)
        return tz_name
    except Exception as e:
        logger.debug(f"Could not determine timezone for {airport.name}: {e}")
        return None


def _approx_timezone_offset(lon: float) -> float:
    return lon / 15.0


def get_timezone_offset(
    airport: Airport,
    dt_utc: Optional[datetime] = None,
) -> Optional[float]:
    tz_name = get_timezone_for_airport(airport)

    if tz_name is None:
        return _approx_timezone_offset(airport.longitude_deg)

    if dt_utc is None:
        dt_utc = datetime.utcnow()

    if dt_utc.tzinfo is None:
        dt_utc = dt_utc.replace(tzinfo=ZoneInfo("UTC"))

    try:
        tz = ZoneInfo(tz_name)
        offset = tz.utcoffset(dt_utc)
        if offset is None:
            return None
        return offset.total_seconds() / 3600.0
    except ZoneInfoNotFoundError:
        logger.debug(f"Timezone {tz_name} not found")
        return None
    except Exception as e:
        logger.debug(f"Error computing timezone offset: {e}")
        return None


def get_timezone_for_code(code: str, code_type: str = "iata") -> Optional[str]:
    if code_type == "iata":
        airport = get_airport_by_iata(code)
    elif code_type == "icao":
        airport = get_airport_by_icao(code)
    else:
        airport = get_airport_by_iata(code) or get_airport_by_icao(code)

    if airport is None:
        return None

    return get_timezone_for_airport(airport)


def local_time_for_airport(
    airport: Airport, dt_utc: Optional[datetime] = None
) -> Optional[datetime]:
    tz_name = get_timezone_for_airport(airport)

    if tz_name is None:
        return None

    if dt_utc is None:
        dt_utc = datetime.utcnow()

    if dt_utc.tzinfo is None:
        dt_utc = dt_utc.replace(tzinfo=ZoneInfo("UTC"))

    try:
        tz = ZoneInfo(tz_name)
        local_dt = dt_utc.astimezone(tz)
        return local_dt
    except ZoneInfoNotFoundError:
        logger.debug(f"Timezone {tz_name} not found")
        return None
    except Exception as e:
        logger.debug(f"Error converting to local time: {e}")
        return None


def local_time_for_code(
    code: str, code_type: str = "iata", dt_utc: Optional[datetime] = None
) -> Optional[datetime]:
    if code_type == "iata":
        airport = get_airport_by_iata(code)
    elif code_type == "icao":
        airport = get_airport_by_icao(code)
    else:
        airport = get_airport_by_iata(code) or get_airport_by_icao(code)

    if airport is None:
        return None

    return local_time_for_airport(airport, dt_utc)



import pytest

from aeronavx.core.emissions_advanced import (
    AircraftType,
    FuelType,
    calculate_flight_emissions,
    compare_saf_savings,
)


def test_calculate_flight_emissions_basic():
    result = calculate_flight_emissions(
        "IST",
        "JFK",
        aircraft_type=AircraftType.WIDE_BODY,
        fuel_type=FuelType.JET_A1,
        load_factor=0.85,
        code_type="iata",
    )

    assert result.distance_km > 0
    assert result.fuel_kg > 0
    assert result.total_co2_kg > 0
    assert result.co2_per_passenger_kg > 0


def test_compare_saf_savings():
    savings = compare_saf_savings(
        "IST",
        "JFK",
        saf_percentage=50.0,
        aircraft_type=AircraftType.WIDE_BODY,
        load_factor=0.85,
        code_type="iata",
    )

    assert savings.co2_reduction_kg > 0
    assert 0 < savings.reduction_percentage < 100


def test_invalid_load_factor():
    with pytest.raises(ValueError):
        calculate_flight_emissions("IST", "JFK", load_factor=1.5)



import pytest

from aeronavx.core.passenger_experience import JetLagSeverity, TravelDirection, calculate_jet_lag
from aeronavx.models.airport import Airport


def _airport(code, lon):
    return Airport(
        id=None,
        ident=code,
        type="large_airport",
        name=f"{code} Airport",
        latitude_deg=0.0,
        longitude_deg=lon,
        elevation_ft=None,
        continent=None,
        iso_country=None,
        iso_region=None,
        municipality=None,
        scheduled_service=None,
        gps_code=code,
        iata_code=code,
        local_code=None,
        home_link=None,
        wikipedia_link=None,
        keywords=None,
    )


def test_calculate_jet_lag_direction_and_severity():
    origin = _airport("AAA", 0.0)
    destination = _airport("BBB", 90.0)

    result = calculate_jet_lag(
        origin,
        destination,
        age=30,
        origin_offset=0.0,
        destination_offset=6.0,
    )

    assert result.timezone_difference_hours == 6.0
    assert result.direction == TravelDirection.EASTWARD
    assert result.severity == JetLagSeverity.SEVERE
    assert result.estimated_recovery_days > 0


def test_jet_lag_age_adjustment():
    origin = _airport("AAA", 0.0)
    destination = _airport("BBB", -60.0)

    younger = calculate_jet_lag(
        origin,
        destination,
        age=18,
        origin_offset=0.0,
        destination_offset=-4.0,
    )
    older = calculate_jet_lag(
        origin,
        destination,
        age=55,
        origin_offset=0.0,
        destination_offset=-4.0,
    )

    assert younger.estimated_recovery_days < older.estimated_recovery_days


def test_jet_lag_invalid_age():
    origin = _airport("AAA", 0.0)
    destination = _airport("BBB", 15.0)

    with pytest.raises(ValueError):
        calculate_jet_lag(origin, destination, age=0, origin_offset=0.0, destination_offset=1.0)
